"""Actor implementations.

Specific actor subclasses maybe should be in a separate module, with just the
abstract stuff here, but we'll see how it goes.
"""

from OpenGL.GL import *
import pygame

import behaviors
import graphics
import resources
import volleys
import window
from vector2d import *

#### Actors ###################################################################

class Actor(object):
    """Represents some entity or behavior on the game world.
    
    Some actors may be visible, but others may simply be behind-the-scenes
    scripts which trigger certain actions. As a result there's no notion of a
    "sprite" in the base class.
    """
    def __init__(self, game, pos, type=None, radius=None, layer=0):
        self.game = game
        self.pos = pos
        self.old_pos = None
        self.is_dead = False
        self.radius = radius
        self.type = type
        self.layer = layer

    def draw(self):
        """Draws the actor to the screen.
        
        The coordinates are already local when this is called, so drawing to
        (0, 0) draws to the position of this actor.

        By default draw nothing.
        """
        pass

    def update(self):
        """Updates the actor.
        
        By default do nothing.
        """
        pass
    
    def kill(self):
        """Requests that the actor die.
        
        If the actor needs to do some cleanup it should override this function.
        Not setting self.is_dead doesn't guarantee that the actor will stay
        around, though, as the Game instance might need to remove actors for
        other reasons.
        """
        self.is_dead = True

class SpriteActor(Actor):
    """Sits there and looks pretty."""
    def __init__(self, game, pos, sprite, layer=0):
        Actor.__init__(self, game, pos, layer=layer)
        self.sprite = sprite
        
    def draw(self):
        self.sprite.draw()

class BehaviorActor(Actor):
    """An actor with a behavior."""
    def __init__(self, game, pos, type=None, radius=None, layer=0):
        Actor.__init__(self, game, pos, type, radius, layer)
        self.behavior_stack = []

    def update(self):
        # Calls the behavior stack to activate the character's action
        if self.behavior_stack:
            self.behavior_stack[-1](self)

    def push_behavior(self, *args):
        # Pushes a reversed list of the arguments onto the behavior stack
        args = list(args)
        args.reverse()
        self.behavior_stack.extend(args)

    def pop_behavior(self):
        self.behavior_stack.pop()

class TextActor(BehaviorActor):
    """An actor for displaying text."""
    def __init__(self, game, pos, text, layer=0, font="vera_serif", align='left'):
        BehaviorActor.__init__(self, game, pos, layer=layer)
        self.text = text
        self.opacity = 1.0
        self.scale = 1.0
        self.font = font
        self.align = align
        self.colour = (0, 0, 0)

    def draw(self):
        glColor(self.colour[0], self.colour[1], self.colour[2], self.opacity)
        graphics.draw_string(resources.font_maps[self.font], self.text,
                             scale=self.scale, align=self.align)
        glColor(1, 1, 1, 1)

class ScriptedActor(BehaviorActor):
    """An actor with scripted behavior."""
    def __init__(self, game, pos, animation_set, type=None, radius=6, layer=0):
        BehaviorActor.__init__(self, game, pos, type, radius, layer)
        self.animation_set = animation_set
        self.vel = Vector2d(0, 0)
        self.__state = "idle"
        self.__state_timer = 0
        self.opacity = 1.0
    
    def update(self):
        self.old_pos = self.pos

        BehaviorActor.update(self)

        # Move according to velocity
        self.pos += self.vel

        # Allow the actor to go 'buffer' pixels out of the screen so that they
        # have a chance to disappear cleanly
        buffer = 10
        width, height = self.game.get_dimensions()
        if self.pos.x < -buffer or width + buffer < self.pos.x or \
           self.pos.y < -buffer or height + buffer < self.pos.y:
            self.kill()

        if self.__state_timer > 0:
            self.__state_timer = self.__state_timer - 1
            if self.__get_animation():
                self.__get_animation().update(1)
            if self.__state_timer == 0:
                self.__state = "idle"
                self.__state_timer = 0
                self.__get_animation().reset()
        else:
            if self.__get_animation():
                self.__get_animation().update(1)

    def draw(self):
        animation = self.__get_animation()
        if animation:
            glColor(1, 1, 1, self.opacity)
            self.__get_animation().draw()
            glColor(1, 1, 1, 1)

    def __get_animation(self):
        if self.animation_set:
            return self.animation_set[self.__state] or \
                   self.animation_set["idle"] or None
        else:
            return None

    def change_state(self, state, time):
        self.__state = state
        self.__state_timer = time
        if self.__get_animation():
            self.__get_animation().reset()

class ScoreWidgetActor(BehaviorActor):
    def __init__(self, game, pos, layer=0):
        BehaviorActor.__init__(self, game, pos, layer=layer)
        self.opacity = 1.0
    def draw(self):
        # Draw score, progress
        score_left = 80
        score_right = 720
        score_top = 584
        score_bottom = 599
        score_min_width = 25
        end_time = 120000

        t = self.game.get_music_pos() / float(end_time)
        current_right = int((score_left + score_min_width) * (1 - t) \
                            + score_right * t)
        if current_right > score_right:
            current_right = score_right
        glColor(0.1, 0.1, 0.1, self.opacity)
        glBegin(GL_QUADS)
        glVertex(score_left, score_bottom)
        glVertex(current_right, score_bottom)
        glVertex(current_right, score_top)
        glVertex(score_left, score_top)
        glEnd()
        glColor(0.5, 0.5, 0.5, self.opacity)
        glBegin(GL_LINE_LOOP)
        glVertex(score_left, score_bottom)
        glVertex(current_right, score_bottom)
        glVertex(current_right, score_top)
        glVertex(score_left, score_top)
        glEnd()
        glColor(0.9, 0.9, 0.9, self.opacity)
        graphics.draw_string(resources.font_maps["vera_serif_small"],
                             str(self.game.get_score()),
                             current_right - 2, score_top-2, align = 'right')
        glColor(1, 1, 1)

class GraphActor(ScriptedActor):
    """An actor that draws a graph."""
    def draw(self):
        glColor(0, 0, 0, self.opacity)
        self.game.log.draw(self.opacity)
        glColor(1, 1, 1, self.opacity)

class PlayerActor(ScriptedActor):
    """An actor that the player controls"""
    def __init__(self, game, pos, animation_set):
        ScriptedActor.__init__(self, game, pos, animation_set, "player", 10)
        self.push_behavior(behaviors.player_control())
    
    def collide(self, bullet):
        #The duration of the hurt animation is 28/60
        self.change_state("hurt", 28)
        self.vel = self.vel + bullet.vel * 4
