from vector2d import *

# Parameters and constants
_width = 800
_height = 600
_small_border = 30
_big_border = 100
_player_border = [
    Vector2d(         _big_border,   _small_border),
    Vector2d(_width - _big_border,   _small_border),
    Vector2d(_width - _small_border,           _big_border),
    Vector2d(_width - _small_border, _height - _big_border),
    Vector2d(_width - _big_border,   _height - _small_border),
    Vector2d(         _big_border,   _height - _small_border),
    Vector2d(_small_border,          _height - _big_border),
    Vector2d(_small_border,                    _big_border)
]

# Behaviors return functions that dictate an actor's behavior (amazing!) that
# should be pushed onto a character's behavior stack via push_behavior.

def wait_for(time):
    """Make the actor wait for the given amount of time in seconds"""
    # Since variables in outer scopes are read-only in python, we need to use a
    # mutable data structure to be able to modify it.
    time_left = [time]
    def behavior(self):
        time_left[0] = time_left[0] - 1
        if time_left[0] <= 0:
            self.pop_behavior()
    return behavior

def wait_until(time):
    """Make the actor wait until the given timestamp"""
    def behavior(self):
        if self.game.get_music_pos() >= time:
            self.pop_behavior()
    return behavior

def player_control():
    """The behavior that is assigned to the player to allow them to move."""
    def behavior(self):
        # controls handling
        new_vel = self.game.get_player_direction() * 4
        self.vel = self.vel * 0.8 + new_vel * 0.2

        # clamp to level
        next_pos = self.vel + self.pos
        for i in range(-len(_player_border), 1):
            a, b = _player_border[i], _player_border[i+1]
            a_to_b = b - a
            a = a + a_to_b.rotate90().get_normalized() * self.radius
            a_to_player = next_pos - a
            if a_to_b.cross(a_to_player) < 0:
                next_pos = a + a_to_player.project(a_to_b)
        self.vel = next_pos - self.pos

        # update the game with our position
        self.game.set_target(self.pos)
        self.game.set_vel(self.vel)
    return behavior

def change_velocity(vel):
    """Makes the actor move in a certain direction."""
    def behavior(self):
        self.vel = vel
        self.pop_behavior()
    return behavior

def change_state(state, time):
    """Change state for a certain amount of time."""
    def behavior(self):
        self.change_state(state, time)
        self.pop_behavior()
    return behavior

def wait_for_confirm():
    """Wait for the player to hit the ok button."""
    def behavior(self):
        if self.game.key_was_pressed("ok"):
            self.pop_behavior()
    return behavior

def spawn_and_wait(*actors_to_spawn):
    """Spawn an actor and wait for it to die."""
    first_time = [True]
    def behavior(self):
        if first_time[0]:
            for actor in actors_to_spawn:
                self.game.add_actor(actor)
            first_time[0] = False
        for actor in actors_to_spawn:
            if not actor.is_dead:
                return
        self.pop_behavior()
    return behavior
    
def wait_until_gone(actor_type):
    def behavior(self):
        if not self.game.is_actor_of_type(actor_type):
            self.pop_behavior()
    return behavior

def fade(increment, actor = None):
    def behavior(self):
        if actor == None:
            target = self
        else:
            target = actor
        target.opacity = min(1.0, max(0.0, target.opacity + increment))
        if target.opacity == 1.0 or target.opacity == 0.0:
            self.pop_behavior()
    return behavior

def kill(actor = None):
    def behavior(self):
        if actor == None:
            self.kill()
        else:
            actor.kill()
        self.pop_behavior()
    return behavior

def run_function(f):
    def behavior(self):
        f(self.game)
        self.pop_behavior()
    return behavior

def home_for(time):
    timer = [time]
    def behavior(self):
        if timer[0] == 0:
            self.pop_behavior()
            return
        timer[0] = timer[0] - 1

        old_vel = self.vel
        speed = old_vel.magnitude()

        if old_vel.is_zero():
            return

        direction = self.game.get_target() - self.pos
        if not direction.is_zero():
            direction = direction.get_normalized()
        self.vel = self.vel + direction / 20

        if self.vel.is_zero():
            self.vel = old_vel
        else:
           self.vel = self.vel.get_normalized() * speed
    return behavior

def text_entry(length):
    prompt = [""]
    timer = [0]
    def behavior(self):
        if self.game.is_string_ready():
            self.pop_behavior()
        elif not self.game.is_string_input_active():
            prompt[0] = self.text
            self.game.start_string_input(length)
        else:
            suffix = int(timer[0] / 20) % 2 == 0 and "_" or ""
            self.text = prompt[0] + self.game.get_string_input()
            if len(self.text) < length :
                self.text += suffix
        timer[0] = timer[0] + 1
    return behavior


def multiplier_popup(mult):
    t = [0]
    def behavior(self):
        if t[0] > 90:
            self.kill()
        else:
            self.scale = 0. + t[0] / 120. * (mult+1)
            self.opacity = 0.5 - t[0] / 180.
            t[0] = t[0] + 1
    return behavior

def controls(follow):
    done = [False]
    def behavior(self):
        self.pos = follow.pos + Vector2d(35, 30)
        if self.game.get_music_pos() >= 1500:
            done[0] = True
        if done[0]:
            self.opacity = self.opacity - (1./60)
        else:
            self.opacity = 1.0
        if self.opacity <= 0:
            self.kill()
    return behavior
