"""Core game logic management.

Handles the frame-to-frame updating, rendering, and event handling.
"""

import math
import random
import sys
import pygame

from pygame.mixer_music import *
from OpenGL.GL import *
from OpenGL.GLU import *

import actor
import behaviors
import graphics
import kernel
import resources
import util
import window

from vector2d import Vector2d
from log import Log

width = 800
height = 600
version= "1.0.0"

# special keywords which can be used instead of the raw key constants
keybindings = {
    "quit": pygame.K_ESCAPE,
    "pause": pygame.K_p,
    "ok": pygame.K_RETURN,
    "debug": pygame.K_d,
    "activate": pygame.K_SPACE,
    "version": pygame.K_v,
    "no_music": pygame.K_RETURN,

    "player_left": pygame.K_LEFT,
    "player_right": pygame.K_RIGHT,
    "player_up": pygame.K_UP,
    "player_down": pygame.K_DOWN,
}

#### Game Class ###############################################################

class Game(object):
    """Manages the overall state of the game.

    It has methods called on it from the main loop. handle_event() is called
    for every pygame event. update() is called to update the game state 60
    times a second, then draw() is called to render everything.

    The game holds a list of actors, which encapsulate in-game objects,
    behaviours, etc.
    """

    def __init__(self, level_init):
        """Initialize the game.

        This does some resource loading as well, so make sure you don't
        construct this object before calling kernel.init().
        """
        # Initialize non-resetable variables
        self.__pressed_keys = set()
        self.__old_pressed_keys = set()
        self.__screen_size = (width, height)
        self.__new_screen_size = self.__screen_size
        self.__camera_center = Vector2d(width/2, height/2)
        self.__music_loaded = False
        self.__level = level_init

        # Load all resources
        resources.load_resources()

        # Initialize game variables
        self.__reset_variables()
        
    def __reset_variables(self):
        """Initialize variables that are reset with every new game"""
        # Initialize the window stack as empty
        self.__window_stack = []

        # Actors control any special behaviour or entity in the world
        self.__actors = []
        
        # Music position and status tracking
        self.__music_succeeded = False
        self.__music_started = False
        self.__music_pos = 0
        
        # Difficulty and score tracking
        self.__difficulty_counter = 0
        self.__difficulty_mod = 1
        self.__current_score = 0
        self.log = Log()

        # Game reset flag
        self.__signal_reset = False

        # String input buffer
        self.__string_input = None
        self.__string_input_length = 0

        # Level initialization
        self.__level(self)

    def handle_event(self, event):
        """Handles pygame events."""

        if event.type == pygame.QUIT or \
           event.type == pygame.KEYDOWN and event.key == keybindings["quit"]:
            kernel.quit()
            return

        # special stuff if we're in string input mode
        if self.__string_input_length > 0 and \
           (event.type == pygame.KEYUP or event.type == pygame.KEYDOWN):
            if event.type == pygame.KEYDOWN and event.unicode:
                char = event.unicode.encode('ascii', 'replace')
                if event.key == pygame.K_RETURN:
                    self.__string_input_length = 0
                elif event.key == pygame.K_BACKSPACE:
                    self.__string_input = self.__string_input[:-1]
                elif char in graphics.supported_characters and \
                     len(self.__string_input) < self.__string_input_length:
                    self.__string_input = self.__string_input + event.unicode.upper()
        # stuff that's not string input mode
        elif event.type == pygame.KEYUP:
            self.__pressed_keys.discard(event.key)
        elif event.type == pygame.KEYDOWN:
            self.__pressed_keys.add(event.key)
        elif event.type == pygame.VIDEORESIZE:
            self.__new_screen_size = event.size

    def update(self, dt):
        """Execute the main game logic update."""
        
        # If there is a dialog open, the game is paused
        is_paused = len(self.__window_stack) > 0
        if self.__music_started:
            if is_paused:
                self.pause_music()
            else:
                self.play_music()
        
        # Update windows in reverse order, that way if the top one removes
        # itself the next one has a chance to respond.
        for win in reversed(self.__window_stack):
            win.update(self, win == self.__window_stack[-1])
            
        # Display the dialog window if the user pressed the dialog test button
        # if not is_paused and self.key_was_pressed("version"):
        #     self.push_window(window.Dialog("Norwegian Wood v: "+ version +"\nPython v: "+sys.version.split()[0]+"\nPygame v: "+ pygame.version.ver +"\nPyOpenGL v: "+ OpenGL.__version__))
            
        # Add the regular pause dialog if the user presses the pause button.
        if not is_paused and self.key_was_pressed("pause"):
            self.push_window(window.Dimmer())
            self.push_window(window.Pause())
            is_paused = True

        # If the game isn't paused, then we update the actors
        if not is_paused:
            if self.__music_loaded:
                if self.__music_started:
                    self.__music_pos = get_pos()
            elif self.__music_started:
                self.__music_pos = self.__music_pos + 1000./60

            # Update all actors
            for actor in self.__actors[:]:
                actor.update()
            self.__do_collision_check()
            self.__remove_dead_actors()
            if self.__music_started and self.is_actor_of_type("enemy"):
                self.__update_score()

        # Overwrite the old key states so that we can tell which keys are
        # pressed/released between this frame and the next frame
        self.__old_pressed_keys = self.__pressed_keys.copy()
        
        # If the signal is given to reset the game, do so
        if self.__signal_reset :
            self.stop_music()
            self.__reset_variables()

    def __do_collision_check(self):
        player = None
        for a in self.__actors:
            if a.type == "player":
                player = a
                break
        
        if player != None:
            for actor in self.__actors:
                if actor.type == "enemy" and \
                   util.actors_collide(player, actor):
                    player.collide(actor)
                    actor.kill()
                    self.__reset_score()
                    break
        
    def load_music(self, file):
        load(file)
        self.__music_loaded = True

    def play_music(self):
        if self.__music_loaded:
            if not self.__music_succeeded and \
               (not self.__music_started or not get_busy()):
                play()
            else:
                self.__music_succeeded = True
                unpause()
        self.__music_started = True

    def pause_music(self):
        if self.__music_loaded:
            pause()
    
    def stop_music(self):
        if self.__music_loaded:
            stop()
    
    def draw(self):
        """Render the game."""

        # Initialization
        self.__init_gl()

        # Drawing in-game objects, transform to world coordinates.
        glPushMatrix()
        cx, cy = self.__camera_center.coords()
        # Make sure we don't go off the map!
        glTranslate(width/2 - cx, height/2 - cy, 0)

        glColor(0.95, 0.95, 0.95)
        glBegin(GL_LINE_STRIP)
        for actor in self.__actors:
            if actor.type == "enemy":
                glVertex(actor.pos.x, actor.pos.y)
        glEnd()
        glColor(1, 1, 1)

        # Draw all the actors
        for actor in self.__actors:
            glPushMatrix()
            glTranslate(actor.pos.x, actor.pos.y, 0)
            actor.draw()
            glPopMatrix()

        # debug drawing goes on top of the foreground
        # if self.is_in_debug_mode():
        #     self.__draw_debug()

        glPopMatrix()
        # Back in screen coordinates.

        # Draw all the open windows
        for win in self.__window_stack:
            win.draw(win == self.__window_stack[-1])

    def __draw_debug(self):
        """Drawing that only happens in debug mode."""
        glColor(1, 0, 0)

        # actor debug drawing
        for actor in self.__actors:
            glPushMatrix()
            glTranslate(actor.pos.x, actor.pos.y, 0)

            # Draw the actor's position
            glBegin(GL_LINES)
            glVertex(-2, -2)
            glVertex(+2, +2)
            glVertex(-2, +2)
            glVertex(+2, -2)
            glEnd()

            # Draw the actor's collision box if it has one
            if actor.radius:
                glBegin(GL_LINE_LOOP)
                x, y = actor.pos.coords()
                r = actor.radius
                glVertex(-r, -r)
                glVertex(-r, +r)
                glVertex(+r, +r)
                glVertex(+r, -r)
                glEnd()
            glPopMatrix()

        # ugly, but totally useful
        import behaviors
        glBegin(GL_LINE_LOOP)
        for v in behaviors._player_border:
            glVertex(v.x, v.y)
        glEnd()
        
        # Draw the performance chart (temporary)
        # glPushMatrix()
        # glTranslate(125, 250, 0)
        # self.log.draw()
        # glPopMatrix()

        # Draw the fps text
        # graphics.draw_string(resources.font_maps["outline"], str(kernel.get_fps()))

        glColor(1, 1, 1)

    def __init_gl(self):
        """Just initializes some OpenGL stuff for rendering.

        Call it before every frame's draw.
        """

        # Resize display if necessary
        if self.__new_screen_size != self.__screen_size:
            # don't let the user resize to smaller than the minimum size
            self.__new_screen_size = (max(self.__new_screen_size[0], width),
                                      max(self.__new_screen_size[1], height))

            # resize the window, taking care not to lose textures
            self.__screen_size = self.__new_screen_size
            graphics.unload_textures()
            kernel.resize(*self.__screen_size)
            graphics.reload_textures()

            # calculate the new scale
            x_scale = self.__screen_size[0] / width
            y_scale = self.__screen_size[1] / height
            scale = int(max(1, min(x_scale, y_scale)))
            assert scale >= 1

            # set the new opengl viewport
            x_padding = (self.__screen_size[0] - width * scale) / 2
            y_padding = (self.__screen_size[1] - height * scale) / 2
            glScissor(x_padding, y_padding, width * scale, height * scale)
            glViewport(x_padding, y_padding, width * scale, height * scale)

        # Clear screen
        glDisable(GL_SCISSOR_TEST)
        glClearColor(0, 0, 0, 0)
        glClear(GL_COLOR_BUFFER_BIT)
        glEnable(GL_SCISSOR_TEST)
        glClearColor(1, 1, 1, 0)
        glClear(GL_COLOR_BUFFER_BIT)

        # Coordinate space is in pixels from the top left
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, width, height, 0, 1, -1)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # Set up default attributes
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def add_actor(self, actor):
        """Appends an actor to the list of actors."""
        i = 0
        while i < len(self.__actors) and self.__actors[i].layer <= actor.layer:
            i = i + 1
        self.__actors.insert(i, actor)

    def __remove_dead_actors(self):
        self.__actors = [a for a in self.__actors if not a.is_dead]
    
    def get_speed_mod(self):
        return (self.__difficulty_mod - 1) * 0.1 + 1
        
    def __update_score(self):
        """Update the current score"""
        self.__update_difficulty()
        self.__current_score += self.__difficulty_mod
    
    def __update_difficulty(self):
        previous_mod = self.__difficulty_mod
        
        self.__difficulty_counter += 1
        if self.__difficulty_counter > 1600 :
            self.__difficulty_mod = 5
        elif self.__difficulty_counter > 900 :
            self.__difficulty_mod = 4
        elif self.__difficulty_counter > 500 :
            self.__difficulty_mod = 3
        elif self.__difficulty_counter > 200 :
            self.__difficulty_mod = 2
        else :
            self.__difficulty_mod = 1
            
        if previous_mod != self.__difficulty_mod :
            if self.__difficulty_mod > 1:
                mult_actor = actor.TextActor(self, Vector2d(30, 30),
                                             "x" + str(self.__difficulty_mod),
                                             font='vera_serif_mult')
                mult_actor.scale = 0.
                mult_actor.push_behavior(
                    behaviors.multiplier_popup(self.__difficulty_mod))
                self.add_actor(mult_actor)
            self.log.record_multiplier(self.__difficulty_mod)
    
    def __reset_score(self):
        self.__difficulty_counter = 0
        self.__current_score = max(self.__current_score - 100, 0)
        self.__update_difficulty
        self.log.record_collision()
        
    def get_hit_count(self):
        return self.log.get_hit_count()
        
    def get_player_direction(self):
        dir = Vector2d((-1 if pygame.K_LEFT  in self.__pressed_keys else 0) +
                       ( 1 if pygame.K_RIGHT in self.__pressed_keys else 0),
                       (-1 if pygame.K_UP    in self.__pressed_keys else 0) +
                       ( 1 if pygame.K_DOWN  in self.__pressed_keys else 0))
        if dir.is_zero():
            return dir
        else:
            return dir.get_normalized()

    def signal_reset(self, level):
        """Signal that the game should reset on the next update."""
        self.__level = level
        self.__signal_reset = True

    def is_in_debug_mode(self):
        return self.key_is_down("debug")
        
    def is_actor_of_type(self, actor_type):
        #Does an actor of the given type exist?
        for actor in self.__actors:
            if actor.type == actor_type:
                return True
        return False

    def __map_key(self, key):
        """Handles transforming keyword strings to their keybindings."""
        if isinstance(key, str):
            assert key in keybindings, "tried to use an undefined key: " + key
            return keybindings[key]
        else:
            return key

    def get_dimensions(self):
        return (width, height)

    def key_is_down(self, key):
        """Returns true if the key is currently down, false otherwise."""
        return self.__map_key(key) in self.__pressed_keys

    def key_was_pressed(self, key):
        """Returns true if the key is down this frame but not the last."""
        real_key = self.__map_key(key)
        return real_key in self.__pressed_keys and \
               real_key not in self.__old_pressed_keys

    def key_was_released(self, key):
        """Returns true if the key was down last frame but not now."""
        real_key = self.__map_key(key)
        return real_key not in self.__pressed_keys and \
               real_key in self.__old_pressed_keys

    def start_string_input(self, length):
        assert length >= 1, "string input length must be at least 1"
        self.__string_input = ""
        self.__string_input_length = length

    def is_string_input_active(self):
        return self.__string_input != None and self.__string_input_length > 0

    def is_string_ready(self):
        return self.__string_input != None and self.__string_input_length == 0

    def get_string_input(self):
        assert self.__string_input != None, "no string available"
        return self.__string_input

    def reset_string_input(self):
        self.__string_input == None

    def set_camera_center(self, pos):
        """Set the target for the game's camera."""
        self.__camera_center = pos
        
    def get_score(self):
        return self.__current_score

    def set_target(self, target):
        """Sets the position that enemies will fire towards."""
        self.__target = target

    def get_target(self):
        """Gets the position that enemies will fire towards."""
        return self.__target
    
    def set_vel(self, vel):
        """Sets the current velocity of the player."""
        self.__vel = vel
        
    def get_vel(self):
        """Gets the current velocity of the player."""
        return self.__vel
    
    def get_music_pos(self):
        return self.__music_pos

    def pop_window(self):
        """Removes the top window from the window stack."""
        self.__window_stack.pop()

    def push_window(self, window):
        """Pushes a new window on top of the window stack."""
        self.__window_stack.append(window)

#### Utility ##################################################################

def try_to_clamp(value, lower, upper):
    """Returns value clamped to the range if possible, otherwise centered."""
    if lower < upper:
        return max(lower, min(value, upper))
    else:
        return lower + upper / 2.
