"""Graphical utilities, such as textures, sprites, fonts."""

import math
import weakref

import pygame
from OpenGL.GL import *
from OpenGL.GLU import *

from vector2d import *

#### Textures #################################################################

class Texture2D(object):
    """Keeps track of a single OpenGL texture."""

    __all_textures = []

    def __init__(self, name=0, pixel_width=0, pixel_height=0):
        """Creates a texture wrapper with the given OpenGL texture name."""
        self.__name = name
        self.pixel_width = pixel_width
        self.pixel_height = pixel_height
        Texture2D.__all_textures.append(weakref.ref(self))

    def bind(self):
        """Binds the texture and enables OpenGL texturing."""
        assert not self.is_empty(), "Tried to bind an empty texture."
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self.__name)

    def unbind(self):
        """Unbinds the texture and disables OpenGL texturing."""
        glBindTexture(GL_TEXTURE_2D, 0)
        glDisable(GL_TEXTURE_2D)

    def delete(self):
        """Deletes the texture from video memory.

        This object is made empty after you call this method. Don't try to bind
        it until you've called set_name on it to fill it again.
        """
        assert not self.is_empty(), "Tried to delete an empty texture."
        glDeleteTextures([self.__name])
        self.__name = 0
        self.pixel_width = 0
        self.pixel_height = 0

    @staticmethod
    def delete_all():
        """Somebody set up us the bomb!"""
        for ref in Texture2D.__all_textures:
            if ref() and not ref().is_empty():
                ref().delete()
        del Texture2D.__all_textures[:]

    def fill(self, name, pixel_width, pixel_height):
        """Fills this object with a new OpenGL texture name.

        This object should be empty before you call this method.
        """
        assert self.is_empty(), "Called set_name on a non-empty texture."
        self.__name = name
        self.pixel_width = pixel_width
        self.pixel_height = pixel_height

    def is_empty(self):
        """Returns false if this contains a valid OpenGL texture."""
        return self.__name == 0

    def __del__(self):
        """Catches textures which fall through the cracks.

        Textures should be deleted manually when possible to free up video
        memory.
        """
        if not self.is_empty():
            self.delete()

_textures = []
_textures_to_reload = []

def _register_texture(texture, texture_initializer):
    _textures.append((texture, texture_initializer))

def unload_textures():
    for texture, texture_initializer in _textures:
        if not texture.is_empty():
            texture.delete()
            _textures_to_reload.append((texture, texture_initializer))

def reload_textures():
    while _textures_to_reload:
        texture, texture_initializer = _textures_to_reload.pop()
        texture_initializer(texture)

def texture_from_file(path):
    """Loads the image given by path into the given texture."""

    def texture_initializer(texture):
        surface = pygame.image.load(path)
        _init_texture_from_surface(texture, surface)

    texture = Texture2D()

    _register_texture(texture, texture_initializer)
    texture_initializer(texture)

    return texture

def _init_texture_from_surface(texture, surface):
    # If the surface is not of power-of-two sizes, correct it.
    surface = ensure_surface_power_of_two(surface)

    name = glGenTextures(1)

    glBindTexture(GL_TEXTURE_2D, name)
    # This makes the textures stay blocky when scaled up.
    glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    # Do some nice filtering on scaled-down textures so that they don't
    # shimmer.
    glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                   GL_LINEAR_MIPMAP_LINEAR)
    # Textures don't wrap around.
    glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
    glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)

    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA,
                      surface.get_width(), surface.get_height(),
                      GL_RGBA, GL_UNSIGNED_BYTE,
                      pygame.image.tostring(surface, "RGBA"))
    glBindTexture(GL_TEXTURE_2D, 0)
    texture.fill(name, surface.get_width(), surface.get_height())

#### Sprites ##################################################################

class Sprite(object):
    """Represents a sprite that you can draw on screen and stuff."""

    def __init__(self, texture, size, origin=(0, 0), texture_area=(0,0,1,1)):
        """Initializes a new sprite object.

        size is a pair of numbers representing the size of the sprite in world
        coordinates. origin is a pair of numbers representing the center of the
        sprite. For example, a sprite with size=(100, 100) and origin=(50, 50)
        would be centered. An origin of (0, 0) means that the top-left of the
        sprite is centered.

        texture_area is a tuple consisting of x, y, width, and height values
        which determine a sub-area of the texture to render onto the sprite.
        These are in OpenGL format, so they count from the top-left of the
        texture, and vary from 0 to 1.
        """
        self.size = size
        self.origin = origin
        self.__texture = texture
        self.__texture_area = texture_area

    def draw(self, x=0, y=0):
        """Draws a sprite offset to the given coordinates."""
        self.__texture.bind()

        size = self.size
        origin = self.origin
        tex_area = self.__texture_area

        glBegin(GL_QUADS)

        glTexCoord(tex_area[0], tex_area[1])
        glVertex(x - origin[0], y - origin[1])

        glTexCoord(tex_area[0] + tex_area[2], tex_area[1])
        glVertex(x + size[0] - origin[0], y - origin[1])

        glTexCoord(tex_area[0] + tex_area[2], tex_area[1] + tex_area[3])
        glVertex(x + size[0] - origin[0],
                 y + size[1] - origin[1])

        glTexCoord(tex_area[0], tex_area[1] + tex_area[3])
        glVertex(x - origin[0], y + size[1] - origin[1])

        glEnd()

        self.__texture.unbind()

#### Animations ###############################################################

class Animation(object):
    """Represents a single linear animation."""

    def __init__(self, frames):
        """Initializes with a set of frames.

        Each frame is a tuple containing a sprite to display, and a number
        representing the duration of that frame. (in seconds)
        """
        assert len(frames) > 0

        self.frames = frames
        self.reset()

    def _current_frame(self):
        return self.frames[self.current_frame]

    def update(self, time):
        """Moves the animation forward the amount of time given."""
        self.time_remaining -= time
        # Change frame if necessary
        while self.time_remaining < 0:
            self.current_frame += 1
            if self.current_frame == len(self.frames):
                self.current_frame = 0
            self.time_remaining += self._current_frame()[1]

    def reset(self):
        self.current_frame = 0
        self.time_remaining = self.frames[0][1]

    def draw(self, x=0, y=0):
        """Draws the current sprite."""
        self._current_frame()[0].draw(x, y)


class AnimationSet(object):
    """Represents a series of animations for 8 directions."""

    def __init__(self, animations):
        """Initializes with all animations present in the input dictionary.

        The keys of the dictionary should be cardinal directions in the form
        of Vector2d instances, and the values should be Animation instances.
        The ZERO direction is the idle animation.
        """
        self.animations = animations

    def __getitem__(self, item):
        """Acts as an indexer for the animations dictionary."""
        return self.animations.get(item)


#### Text #####################################################################

# The characters we render into font maps (+1 at the end to include the ~)
supported_characters = [chr(c) for c in range(ord(" "), ord("~")+1)]
# The character we use when a character lookup fails
_fallback_character = "?"
# Space between letters in the font map texture
_letter_padding = 1

def make_fontmap(path, size, chars=supported_characters):
    """Creates a fontmap (for use with draw_string()) from a pygame font."""

    def texture_initializer(texture, sprite_rects_buffer=None):
        font = pygame.font.Font(path, size)
        linesize = font.get_linesize()
        # The buffer we render onto
        # There's a bug in the algorithm, a dirty fix is to make sure that the
        # initial surface height is larger than the max line height
        surface = pygame.Surface((512, 128), flags=pygame.SRCALPHA)
        # Current render position
        x = _letter_padding
        y = _letter_padding
        # Render each character into the surface, recording its position
        for c in chars:
            letter = font.render(c, True, (255, 255, 255))
            # If we have no more room on this line go to the next
            if x + letter.get_width() + _letter_padding > surface.get_width():
                x = _letter_padding
                y += linesize + _letter_padding
                assert x + letter.get_width() + _letter_padding <= \
                       surface.get_width(), \
                       "A font letter is too wide to fit on the texture."
                # If we ran out of room in the surface expand it
                if y + linesize + _letter_padding > surface.get_height():
                    new_surface = pygame.Surface(
                        (surface.get_width(), surface.get_height() * 2),
                        flags=pygame.SRCALPHA)
                    new_surface.blit(surface, (0, 0))
                    surface = new_surface
            # Render the texture onto the surface
            surface.blit(letter, (x, y))
            # Save its position
            if sprite_rects_buffer != None:
                sprite_rects[c] = \
                    (x, y, letter.get_width(), letter.get_height())
            # Move to the next character's position
            x += letter.get_width() + _letter_padding
        _init_texture_from_surface(texture, surface)

    # Generate the texture
    sprite_rects = {}
    texture = Texture2D()
    _register_texture(texture, texture_initializer)
    texture_initializer(texture, sprite_rects)

    # Generate the sprites
    sprites = {}
    for c, rect in sprite_rects.items():
        texture_area = normalize_texture_area(texture, rect)
        sprites[c] = Sprite(texture, (rect[2], rect[3]),
                            texture_area=texture_area)
    return sprites

def make_bitmap_fontmap(font_image_path, width, height,
                        chars=supported_characters):
    """Creates a fontmap (for use with draw_string()) from a bitmap font
    surface."""

    def texture_initializer(texture, sprite_rects_buffer=None):
        font_surface = pygame.image.load(font_image_path)

        # Keeps track of letter locations
        # Current render position
        x = 0
        y = 0
        # Calculate the size of the buffer
        num_cols = font_surface.get_width() / width
        num_rows = font_surface.get_height() / height
        surface_width = font_surface.get_width() + num_cols*_letter_padding
        surface_height = font_surface.get_height() + num_rows*_letter_padding
        # The buffer we render onto
        surface = pygame.Surface((surface_width, surface_height),
                flags=pygame.SRCALPHA)
        # Render each character into the surface, recording its position
        for c in chars:
            # Calculate the position of the character on the buffer
            xpos = x*width + x*_letter_padding
            ypos = y*height + y*_letter_padding
            # Render the texture onto the buffer from the font_surface
            surface.blit(font_surface,
                         (xpos+_letter_padding,ypos+_letter_padding),
                         (x*width, y*height, width, height))
            # Save its position
            if sprite_rects_buffer != None:
                sprite_rects_buffer[c] = \
                    (xpos, ypos, width+_letter_padding, height+_letter_padding)
            # Move to the next character
            x += 1
            if x == num_cols:
                x = 0
                y += 1
        _init_texture_from_surface(texture, surface)

    # Generate the texture
    sprite_rects = {}
    texture = Texture2D()
    _register_texture(texture, texture_initializer)
    texture_initializer(texture, sprite_rects)

    # Generate the sprites
    sprites = {}
    for c, rect in sprite_rects.items():
        texture_area = normalize_texture_area(texture, rect)
        sprites[c] = Sprite(texture, (rect[2], rect[3]),
                            texture_area=texture_area)
    return sprites

def draw_string(fontmap, string, xOrigin=0, yOrigin=0, scale=1, align='left'):
    """Renders a string on-screen using a fontmap."""
    x = 0
    y = 0
    width = 0
    for c in string:
        if c == "\n":
            x = 0
        else:
            sprite = fontmap.get(c, False) or fontmap[_fallback_character]
            x = x + sprite.size[0]
        width = max(width, x)

    x = 0
    y = 0
    glPushMatrix()
    glTranslate(xOrigin, yOrigin, 0)
    glScale(scale, scale, scale)
    if align == 'right':    glTranslate(-width, 0, 0)
    elif align == 'center': glTranslate(-width/2, 0, 0)
    else:                   assert align == 'left'
    for c in string:
        if c == "\n":
            x = xOrigin
            y = y + fontmap[_fallback_character].size[1]
        else:
            sprite = fontmap.get(c, False) or fontmap[_fallback_character]
            sprite.draw(x, y)
            x = x + sprite.size[0]
    glPopMatrix()

#### General Surface Utilities ################################################

def ensure_surface_power_of_two(surface):
    """Returns a power-of-two surface with the same contents as the argument.

    If the source surface is already power-of-two, it is returned immediately.
    The surface is not scaled, and when a copy is made, the unused portion of
    the surface is left transparent.
    """
    if _is_power_of_two(surface.get_width()) and \
       _is_power_of_two(surface.get_height()):
        return surface

    width = _next_power_of_two(surface.get_width())
    height = _next_power_of_two(surface.get_height())

    corrected_surface = pygame.Surface((width, height), pygame.SRCALPHA)
    corrected_surface.fill(pygame.Color(0, 0, 0, 0))

    # TODO: extend throughout the rest of the texture GL_CLAMP-style
    corrected_surface.blit(surface, (0, 0))
    return corrected_surface

def normalize_texture_area(texture, area):
    """Converts area coordinates from pixels to [0, 1] based on the surface."""
    # We want floats for the next division
    width = float(texture.pixel_width)
    height = float(texture.pixel_height)

    return (area[0] / width, area[1] / height,
            area[2] / width, area[3] / height)

#### Utility ##################################################################

def _next_power_of_two(x):
    return int(2**(math.ceil(math.log(x, 2))))

def _is_power_of_two(x):
    return x == _next_power_of_two(x)
