"""Implements the game's main loop.

First call init() with the window title and resolution, then call
start_main_loop() with a Scene object. Calling quit() will abort the main loop. 
"""

import sys
import pygame
from OpenGL.GL import *

# Call the scene's update function this many times per second. The video redraw
# rate might be lower if frames are skipped.
FPS = 60

# Very short sleeps tend to overshoot by a lot, so if the time left is less
# than this amount, busy-wait instead. Specifically, this should be at least
# 0.01 for Windows, since it has a particularly bad timer.
MIN_SLEEP_TIME = 0.002
if(sys.platform == "win32" or sys.platform == "cygwin"):
    MIN_SLEEP_TIME = 0.01
# pygame.time.wait sometimes sleeps longer than you tell it to, so ask it to
# sleep this much less than we actually want.
SLEEP_HEADROOM = 0.002

_quit = False

# This keeps track of the times of the last few frames, and is used for fps
# measurement. If the number of elements is 3*k+1, the fps display will be
# smooth when it's at 60fps :3
_frame_times = [-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0]

def resize(width, height):
    """Resizes the window for a new resolution.
    
    The original resolution is still used as the coordinate space. This
    function is called on startup, and should also be called in response to
    resize events.
    """
    pygame.display.set_mode(
        (width, height),
        pygame.HWSURFACE | pygame.OPENGL | pygame.DOUBLEBUF | pygame.RESIZABLE)
    glViewport(0, 0, width, height)

def _get_current_time():
    """Returns the current time in seconds."""
    return pygame.time.get_ticks() / 1000.0

def _wait_until(time):
    """Waits until the given time, which is given in seconds."""
    time_to_sleep = time - _get_current_time() - SLEEP_HEADROOM
    if time_to_sleep >= MIN_SLEEP_TIME:
        pygame.time.wait(int(1000 * time_to_sleep))
    # add a small factor (less than 1ms) to eliminate floating point accuracy
    # problems
    while _get_current_time() < time + 0.0001:
        pass

def init(name, resolution):
    """Initializes pygame with a window of the given name and resolution.
    
    Call this only once, and before calling start_main_loop()."""
    pygame.mixer.pre_init(44100, 0, 2, 1024)
    pygame.init()
    resize(*resolution)
    pygame.display.set_caption(name)

def quit():
    """Signals the main loop to end."""
    global _quit
    _quit = True

def start_main_loop(scene):
    """Enters the main loop, using the given scene object for callbacks.
    
    This function only returns after quit() is called."""
    global _quit
    global _frame_times
    UPDATE_TIME = 1.0 / FPS
    last_update_time = _get_current_time()

    while True:
        # wait until it's time to update again
        _wait_until(last_update_time + UPDATE_TIME)
        # handle events
        for event in pygame.event.get():
            scene.handle_event(event)
            if(_quit): return
        # update
        current_time = _get_current_time()
        while last_update_time + UPDATE_TIME <= current_time:
            last_update_time += UPDATE_TIME
            scene.update(UPDATE_TIME)
            if(_quit): return
        # draw
        scene.draw()
        pygame.display.flip()
        # record this frame's time for fps calculations
        _frame_times.append(current_time)
        _frame_times = _frame_times[1:len(_frame_times)]

def get_fps():
    """Gets an estimate of the current FPS."""
    delta = _frame_times[len(_frame_times) - 1] - _frame_times[0]
    return round((len(_frame_times) - 1) / delta)
