"""Provides resource access and loading methods."""

import glob
import os
import pygame
import re
import shutil
import sys

from resources_specs import *
from vector2d import *

import graphics

# The resource dictionaries.
sprites = {}
animations = {}
animation_sets = {}
image_data = {}
font_maps = {}

# From <http://py2exe.org/index.cgi/WhereAmI>, then modified
def module_path():
    if sys.platform != "darwin" and hasattr(sys, "frozen"):
        return os.path.dirname(unicode(sys.executable, sys.getfilesystemencoding( )))
    return "."

# Resource directories
_resource_path = os.path.join(module_path(), "resources")
if not os.path.isdir(_resource_path):
    # hack to make it run from the source directory
    _resource_path = os.path.join("..", _resource_path)
_sprite_path = os.path.join(_resource_path, "sprites")

def load_resources():
    """Loads all specified resources."""
    _load_sprites()
    _load_animations()
    _load_animation_sets()
    _load_fonts()
    _load_bitmap_fonts()

def _load_sprites():
    """Loads all specified sprites."""
    # For each specified spritesheet
    for spritesheet_filename, spritesheet_spec in spritesheets_specs.items():
        # Load its texture into a pygame surface (ensuring it's PoT)
        path = os.path.join(_sprite_path, spritesheet_filename)

        # Create and register the texture
        spritesheet = graphics.texture_from_file(path)

        # For each sprite in this sheet
        for sprite_name, sprite_spec in spritesheet_spec.items():
            # If there's no origin, assume (0, 0)
            size = (abs(sprite_spec[2]), abs(sprite_spec[3]))
            origin = sprite_spec[4:6] if len(sprite_spec) == 6 else (0, 0)
            # Take the area to [0-1] texture space (spec is pixels)
            area = graphics.normalize_texture_area(
                spritesheet,
                sprite_spec[0:4])

            # Create and register the sprite
            sprite = graphics.Sprite(spritesheet, size, origin, area)
            sprites[sprite_name] = sprite

def _load_animations():
    """Loads all specified animations."""
    # For each specified animation
    for animation_name, animation_spec in animations_specs.items():
        # Register its frames
        frames = []
        for frame in animation_spec:
            frames.append((sprites[frame[0]], frame[1]))

        # Create and register the animation
        animation = graphics.Animation(frames)
        animations[animation_name] = animation

def _load_animation_sets():
    """Loads all specified animation sets."""
    # For each specified set
    for set_name, set_spec in animation_sets_specs.items():
        set_anims = {}
        for k, v in set_spec.items():
            set_anims[k] = animations[v]

        # Create and register the set
        set = graphics.AnimationSet(set_anims)
        animation_sets[set_name] = set

def _load_image_data():
    """Loads all image data arrays.
    The images are loaded as two-dimensional arrays, with each pixel stored as
    a 0xRRGGBBAA integer. Access pixels like data[x][y].
    """
    # For each image data spec
    for image_data_name, filename in image_data_specs.items():
        # Load the image
        path = os.path.join(_resource_path, filename)
        image = pygame.image.load(path)
        str = pygame.image.tostring(image, "RGBA")

        # Put the data into a two-dimensional array
        data = []
        for i in range(image.get_width()):
            data.append([])
            for j in range(image.get_height()):
                base = (i + j * image.get_width()) * 4
                data[i].append((ord(str[base + 0]) << 24) +
                               (ord(str[base + 1]) << 16) +
                               (ord(str[base + 2]) <<  8) +
                               (ord(str[base + 3]) <<  0))

        # Register the data
        image_data[image_data_name] = data

def _load_bitmap_fonts():
    """Loads all specified bitmap fonts."""
    for font_name, font_spec in bitmap_font_specs.items():
        path = os.path.join(_resource_path, font_spec[0])
        font_maps[font_name] = \
            graphics.make_bitmap_fontmap(path, font_spec[1], font_spec[2])

def _load_fonts():
    """Loads all specified fonts."""
    for font_name, font_spec in font_specs.items():
        path = os.path.join(_resource_path, font_spec[0])
        if len(font_spec) > 2:
            font_maps[font_name] = \
                graphics.make_fontmap(path, font_spec[1], font_spec[2])
        else:
            font_maps[font_name] = graphics.make_fontmap(path, font_spec[1])

def get_music_file():
    """Return the music file from the resources folder."""
    files = glob.glob(os.path.join(_resource_path, '*.[mM][pP]3'))
    if len(files) == 0:
        files = glob.glob(os.path.join(_resource_path, '*.[oO][gG][gG]'))

    if len(files) == 0:
        return None
    else:
        return files[0]

def copy_file(src):
    """Copy the given file to the resources folder."""
    mp3 = re.compile('.*\.[mM][pP]3')
    ogg = re.compile('.*\.[oO][gG][gG]')
    if mp3.match(src) :
        shutil.copyfile(src, os.path.join(_resource_path, "music.mp3"))
    elif ogg.match(src) :
        shutil.copyfile(src, os.path.join(_resource_path, "music.ogg"))
