"""Scripts to handle level initialization and behaviour.

When this gets non-trivial we'll probably want to split it into several files
in a 'scripts' subdirectory.
"""

import pygame
import os
import actor
import game
import behaviors
import resources
import util
import volleys
import scores
import pickle
import base64
import httplib
import urllib
from vector2d import Vector2d

center = Vector2d(game.width / 2, game.height / 2)
animations = resources.animation_sets

def spawner_behavior(timings):
    # timings is a list of (time, volley_name) tuples
    index = [0]
    def behavior(self):
        i = index[0]
        if i == len(timings):
            self.pop_behavior()
        else:
            self.push_behavior(
                behaviors.wait_until(timings[i][0]),
                behaviors.change_state("fire", 6),
                fire_volley(timings[i][1]))
            index[0] = i + 1
    return behavior

def fire_volley(name):
    def behavior(self):
        getattr(volleys, name)(self.game, self.pos)
        self.pop_behavior()
    return behavior
    
def signal_game_reset(game):
    """ Signal the game to reset at the next update."""
    game.signal_reset(reset_level)
    
def add_border(game):
    """ Add the map border."""
    game.add_actor(actor.SpriteActor(
        game, Vector2d(0,0), resources.sprites["map"], layer=2))

def init_title(game):
    splash = actor.ScriptedActor(game, center, animations["no_fun"])
    splash.opacity = 0.0
    splash.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_for(120),
        behaviors.fade(-2./60),
        behaviors.kill())

    name = actor.ScriptedActor(game, center, animations["norwegian_wood"])
    name.opacity = 0.0
    name.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_for(120),
        behaviors.fade(-2./60),
        behaviors.kill())

    # these two behaviors implement the music loader prompt when no music is
    # found
    def load_music_behavior(self):
        # look for music
        music_file = resources.get_music_file()
        if music_file:
            # if it's found, load it and we're done
            self.pop_behavior()
            game.load_music(music_file)
        else:
            # if it's not found, fade in and prompt the user
            self.pop_behavior()
            instructions.push_behavior(
                behaviors.fade(2./60),
                music_prompt_behavior)
    def music_prompt_behavior(self):
        if self.game.key_was_pressed("no_music"):
            # user doesn't care about music, we're done
            self.push_behavior(
                behaviors.fade(-2./60),
                behaviors.kill())
        elif self.game.key_was_pressed("ok"):
            # user wants to try again, fade out and look for music
            self.pop_behavior()
            self.push_behavior(
                behaviors.fade(-2./60),
                load_music_behavior)

    instructions = actor.ScriptedActor(game, center, animations["instructions"])
    instructions.opacity = 0.0
    instructions.push_behavior(
        load_music_behavior,
        behaviors.kill())

    add_border(game)

    # The title screens in sequence
    title = actor.ScriptedActor(game, center, None)
    title.push_behavior(
        behaviors.wait_for(20),
        behaviors.spawn_and_wait(splash),
        behaviors.spawn_and_wait(name),
        behaviors.spawn_and_wait(instructions),
        behaviors.run_function(init_level),
        behaviors.kill())
    game.add_actor(title)

def init_level(game):
    # The last note, approximately
    last_note = 120000
    
    # The player-controlled actor
    pc = actor.PlayerActor(game, center, animations["john"])
    game.add_actor(pc)

    # One word instructions
    dodge = actor.ScriptedActor(game, center, animations["dodge"])
    dodge.push_behavior(behaviors.controls(pc));
    dodge.opacity = 0.
    game.add_actor(dodge)

    score = actor.ScoreWidgetActor(game, Vector2d(0, 0), layer=3)
    score.push_behavior(
        behaviors.wait_until(last_note),
        behaviors.wait_until_gone("enemy"),
        behaviors.fade(-2./60))
    game.add_actor(score)

    # The four bullet sources
    bullet_spawn_guitar = actor.ScriptedActor(
        game, Vector2d(750, 550), animations["guitar"], layer=3)
    bullet_spawn_guitar.push_behavior(
        spawner_behavior(timings_guitar),
        behaviors.wait_until(last_note),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(bullet_spawn_guitar)
    
    bullet_spawn_bass = actor.ScriptedActor(
        game, Vector2d(750, 50), animations["bass"], layer=3)
    bullet_spawn_bass.push_behavior(
        spawner_behavior(timings_bass),
        behaviors.wait_until(last_note),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(bullet_spawn_bass)
    
    bullet_spawn_sitar = actor.ScriptedActor(
        game, Vector2d(50, 50), animations["sitar"], layer=3)
    bullet_spawn_sitar.push_behavior(
        spawner_behavior(timings_sitar),
        behaviors.wait_until(last_note),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(bullet_spawn_sitar)
    
    bullet_spawn_vocal = actor.ScriptedActor(
        game, Vector2d(50, 550), animations["vocal"], layer=3)
    bullet_spawn_vocal.push_behavior(
        spawner_behavior(timings_vocal),
        behaviors.wait_until(last_note),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(bullet_spawn_vocal)
    
    # End of game
    end_handler = actor.BehaviorActor(game, Vector2d(0, 0))
    end_handler.push_behavior(
        behaviors.wait_until(last_note),
        behaviors.wait_until_gone("enemy"),
        behaviors.fade(-2./60, pc),
        behaviors.run_function(init_end))
    game.add_actor(end_handler)

    # Start the music now!
    game.play_music()

def reset_level(game):
    """ Load the border and reset the level. """
    add_border(game)
    init_level(game)

def init_end(game):
    axis = actor.ScriptedActor(game, Vector2d(120, 70), animations["axis"])
    axis.opacity = 0.0
    axis.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_until_gone("TextEntry"),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(axis)
    
    graph = actor.GraphActor(game, Vector2d(170, 270), None)
    graph.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_until_gone("TextEntry"),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(graph)
    
    score_header = actor.ScriptedActor(game, Vector2d(170, 320), animations["score"])
    score_header.opacity = 0.0
    score_header.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_until_gone("TextEntry"),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(score_header)
    
    score_display = actor.TextActor(game, Vector2d(170, 340), str(game.get_score()), font="vera_serif_big")
    score_display.opacity = 0.0
    score_display.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_until_gone("TextEntry"),
        behaviors.fade(-2./60),
        behaviors.run_function(init_scores))
    game.add_actor(score_display)
    
    name_prompt = actor.ScriptedActor(game, Vector2d(170, 400), animations["enter_name"])
    name_prompt.opacity = 0.0
    name_prompt.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_until_gone("TextEntry"),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(name_prompt)

    name_entry = actor.TextActor(game, Vector2d(170, 430), "", font="vera_serif_big")
    name_entry.opacity = 0.0
    name_entry.type = "TextEntry"
    name_entry.push_behavior(
        behaviors.fade(2./60),
        behaviors.text_entry(10),
        behaviors.fade(-5./60),
        behaviors.kill())
    game.add_actor(name_entry)
    
    press_enter = actor.ScriptedActor(game, Vector2d(400, 510), animations["press_enter"])
    press_enter.opacity = 0.0
    press_enter.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_until_gone("TextEntry"),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(press_enter)

def init_scores(game):
    #Record current score locally
    local_data = scores.record_score(game.get_string_input(), game.get_score(), game.get_hit_count())
    
    #Send current score to server
    global_data = scores.send_score(game.get_string_input(), game.get_score(), game.get_hit_count())
    
    #All actors for the local and global scores
    local_score_screen = []
    global_score_screen = []
    
    #Local score title
    local_score_title = actor.ScriptedActor(game, Vector2d(400, 150), animations["local_scores"], layer=2)
    local_score_title.opacity = 0.0
    local_score_title.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_for_confirm(),
        behaviors.fade(-2./60),
        behaviors.kill())
    local_score_screen.append(local_score_title)

    #Local names and scores
    score_height = 210
    position = 1
    self_score = False
    for high_score in local_data.high_scores:
        rank = actor.TextActor(game, Vector2d(290, score_height), str(position), layer=2, align='right')
        rank.opacity = 0.0
        rank.push_behavior(
            behaviors.fade(2./60),
            behaviors.wait_for_confirm(),
            behaviors.fade(-2./60),
            behaviors.kill()) 
        
        name = actor.TextActor(game, Vector2d(300, score_height), str(high_score[1]), layer=2)
        name.opacity = 0.0
        name.push_behavior(
            behaviors.fade(2./60),
            behaviors.wait_for_confirm(),
            behaviors.fade(-2./60),
            behaviors.kill())        
        
        score = actor.TextActor(game, Vector2d(520, score_height), str(high_score[0]), layer=2, align='right')
        score.opacity = 0.0
        score.push_behavior(
            behaviors.fade(2./60),
            behaviors.wait_for_confirm(),
            behaviors.fade(-2./60),
            behaviors.kill())
        
        if not self_score and str(high_score[0]) == str(game.get_score()) and str(high_score[1]) == game.get_string_input() :
            rank.font = "vera_serif_bold"
            name.font = "vera_serif_bold"
            score.font = "vera_serif_bold"
            self_score = True
        else:
            rank.colour = (0.2, 0.2, 0.2)
            name.colour = (0.2, 0.2, 0.2)
            score.colour = (0.2, 0.2, 0.2)
        
        local_score_screen.append(rank)
        local_score_screen.append(name)
        local_score_screen.append(score)
        score_height += 20
        position += 1
        
    #Display player's score if not in the top 10
    if not self_score:
        gap = actor.TextActor(game, Vector2d(290, score_height - 5), "...", layer=2)
        gap.opacity = 0.0
        gap.push_behavior(
            behaviors.fade(2./60),
            behaviors.wait_for_confirm(),
            behaviors.fade(-2./60),
            behaviors.kill())
        
        score_height += 20
        
        name = actor.TextActor(game, Vector2d(300, score_height), str(game.get_string_input()), layer=2)
        name.opacity = 0.0
        name.font = "vera_serif_bold"
        name.push_behavior(
            behaviors.fade(2./60),
            behaviors.wait_for_confirm(),
            behaviors.fade(-2./60),
            behaviors.kill())    
        
        score = actor.TextActor(game, Vector2d(520, score_height), str(game.get_score()), layer=2, align='right')
        score.opacity = 0.0
        score.font = "vera_serif_bold"
        score.push_behavior(
            behaviors.fade(2./60),
            behaviors.wait_for_confirm(),
            behaviors.fade(-2./60),
            behaviors.kill())
            
        local_score_screen.append(gap)
        local_score_screen.append(name)
        local_score_screen.append(score)
    
    #Global score title
    global_score_title = actor.ScriptedActor(game, Vector2d(400, 150), animations["global_scores"], layer=2)
    global_score_title.opacity = 0.0
    global_score_title.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_for_confirm(),
        behaviors.fade(-2./60),
        behaviors.kill())
    global_score_screen.append(global_score_title)
    
    #Global names and scores
    if global_data[0] > 0 :
        score_height = 210
        position = 1
        self_score = False
        for high_score in global_data[1]:
            rank = actor.TextActor(game, Vector2d(290, score_height), str(position), layer=2, align='right')
            rank.opacity = 0.0
            rank.push_behavior(
                behaviors.fade(2./60),
                behaviors.wait_for_confirm(),
                behaviors.fade(-2./60),
                behaviors.kill()) 
        
            name = actor.TextActor(game, Vector2d(300, score_height), str(high_score[0]).upper(), layer=2)
            name.opacity = 0.0
            name.push_behavior(
                behaviors.fade(2./60),
                behaviors.wait_for_confirm(),
                behaviors.fade(-2./60),
                behaviors.kill())        
        
            score = actor.TextActor(game, Vector2d(520, score_height), str(high_score[1]), layer=2, align='right')
            score.opacity = 0.0
            score.push_behavior(
                behaviors.fade(2./60),
                behaviors.wait_for_confirm(),
                behaviors.fade(-2./60),
                behaviors.kill())
        
            if not self_score and str(high_score[1]) == str(game.get_score()) and str(high_score[0]) == game.get_string_input() :
                rank.font = "vera_serif_bold"
                name.font = "vera_serif_bold"
                score.font = "vera_serif_bold"
                self_score = True
            else:
                rank.colour = (0.2, 0.2, 0.2)
                name.colour = (0.2, 0.2, 0.2)
                score.colour = (0.2, 0.2, 0.2)
        
            global_score_screen.append(rank)
            global_score_screen.append(name)
            global_score_screen.append(score)
            score_height += 20
            position += 1
        
            #If the score is not in the top 10
            if global_data[0] > 10 and position == 7 :
                gap = actor.TextActor(game, Vector2d(290, score_height - 5), "...", layer=2)
                gap.opacity = 0.0
                gap.push_behavior(
                    behaviors.fade(2./60),
                    behaviors.wait_for_confirm(),
                    behaviors.fade(-2./60),
                    behaviors.kill())
                global_score_screen.append(gap)
                score_height += 20
                position = global_data[0] - 1
                
    # Error returned from server
    else:
        error = actor.TextActor(game, Vector2d(260, 210), "Unable to reach score server: Error " + str(global_data[1]), layer=2)
        error.opacity = 0.0
        error.push_behavior(
            behaviors.fade(2./60),
            behaviors.wait_for_confirm(),
            behaviors.fade(-2./60),
            behaviors.kill())
        global_score_screen.append(error)
    
    
    #Credits and game reset screen
    credits = actor.ScriptedActor(game, center, animations["credits"], layer=3)
    credits.opacity = 0.0
    credits.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_for_confirm(),
        behaviors.fade(-2./60),
        behaviors.run_function(signal_game_reset),
        behaviors.kill())
    
    #The "press enter" prompt
    press_enter = actor.ScriptedActor(game, Vector2d(400, 510), animations["press_enter"])
    press_enter.opacity = 0.0
    press_enter.push_behavior(
        behaviors.fade(2./60),
        behaviors.wait_for_confirm(),
        behaviors.fade(-2./60),
        behaviors.fade(2./60),
        behaviors.wait_for_confirm(),
        behaviors.fade(-2./60),
        behaviors.fade(2./60),
        behaviors.wait_for_confirm(),
        behaviors.fade(-2./60),
        behaviors.kill())
    game.add_actor(press_enter)
    
    #The ending screens in sequence
    end = actor.ScriptedActor(game, center, None)
    end.push_behavior(
        behaviors.spawn_and_wait(*local_score_screen),
        behaviors.spawn_and_wait(*global_score_screen),
        behaviors.spawn_and_wait(credits),
        behaviors.run_function(signal_game_reset))
    game.add_actor(end)

timings_vocal = [
    (16379, "targeted"), 
    (17331, "targeted"),
    (17635, "targeted"),
    (17963, "targeted"),
    (18155, "targeted"),
    (19291, "targeted"),
    (19627, "targeted"),
    (19891, "targeted"),
    (20291, "targeted"),
    (21267, "targeted"),
    (21595, "targeted"),
    (21907, "targeted"),
    (22243, "targeted"),
    (24195, "targeted"),
    (25139, "targeted"),
    (25451, "targeted"),
    (25794, "targeted"),
    (26163, "targeted"),
    (27155, "targeted"),
    (27467, "targeted"),
    (27803, "targeted"),
    (28171, "targeted"),
    (29131, "targeted"),
    (29475, "targeted"),
    (29788, "targeted"),
    (30155, "targeted"),
    (31786, "targeted"),
    (32154, "targeted"),
    (32483, "targeted"),
    (32802, "targeted"),
    (33155, "targeted"),
    (33499, "targeted"),
    (33859, "targeted"),
    (34194, "targeted"),
    (34515, "targeted"),
    (34867, "targeted"),
    (35179, "targeted"),
    (35347, "targeted"),
    (35811, "targeted"),
    (35995, "targeted"),
    (39827, "targeted"),
    (40179, "targeted"),
    (40507, "targeted"),
    (40843, "targeted"),
    (41171, "targeted"),
    (41515, "targeted"),
    (41763, "targeted"),
    (42171, "targeted"),
    (42506, "targeted"),
    (42802, "targeted"),
    (43130, "targeted"),
    (43299, "targeted"),
    (43779, "targeted"),
    (43971, "targeted"),
    (48090, "targeted"),
    (49075, "targeted"),
    (49387, "targeted"),
    (49755, "targeted"),
    (50106, "targeted"),
    (51098, "targeted"),
    (51426, "targeted"),
    (51771, "targeted"),
    (52115, "targeted"),
    (53115, "targeted"),
    (53443, "targeted"),
    (53763, "targeted"),
    (54115, "targeted"),
    (56114, "targeted"),
    (57139, "targeted"),
    (57443, "targeted"),
    (57794, "targeted"),
    (58130, "targeted"),
    (59219, "targeted"),
    (59523, "targeted"),
    (59842, "targeted"),
    (60194, "targeted"),
    (61187, "targeted"),
    (61514, "targeted"),
    (61834, "targeted"),
    (62066, "targeted"),
    (80130, "targeted"),
    (80451, "targeted"),
    (80794, "targeted"),
    (81130, "targeted"),
    (81498, "targeted"),
    (81883, "targeted"),
    (82259, "targeted"),
    (82451, "targeted"),
    (82899, "targeted"),
    (83227, "targeted"),
    (83594, "targeted"),
    (83931, "targeted"),
    (84242, "targeted"),
    (84450, "targeted"),
    (88298, "targeted"),
    (88635, "targeted"),
    (88978, "targeted"),
    (89323, "targeted"),
    (89698, "targeted"),
    (89891, "targeted"),
    (90362, "targeted"),
    (90698, "targeted"),
    (91042, "targeted"),
    (91722, "targeted"),
    (91931, "targeted"),
    (92298, "targeted"),
    (92523, "targeted"),
    (96762, "targeted"),
    (97794, "targeted"),
    (98106, "targeted"),
    (98450, "targeted"),
    (98786, "targeted"),
    (99794, "targeted"),
    (100123, "targeted"),
    (100451, "targeted"),
    (100714, "targeted"),
    (101866, "targeted"),
    (102179, "targeted"),
    (102530, "targeted"),
    (102866, "targeted"),
    (104898, "targeted"),
    (105962, "targeted"),
    (106275, "targeted"),
    (106634, "targeted"),
    (106971, "targeted"),
    (108003, "targeted"),
    (108347, "targeted"),
    (108691, "targeted"),
    (109042, "targeted"),
    (110074, "targeted"),
    (110426, "targeted"),
    (110731, "targeted"),
    (111074, "targeted"),
]

timings_sitar = [
    (8406, "targeted_spread"),
    (9441, "targeted_spread"),
    (9756, "targeted_spread"),
    (10071, "targeted_spread"),
    (10406, "targeted_spread"),
    (11401, "targeted_spread"),
    (11756, "targeted_spread"),
    (12071, "targeted_spread"),
    (12386, "targeted_spread"),
    (13331, "targeted_spread"),
    (13696, "targeted_spread"),
    (14011, "targeted_spread"),
    (14346, "targeted_spread"),
    (23166, "targeted_spread"),
    (23506, "targeted_spread"),
    (23861, "targeted_spread"),
    (24201, "targeted_spread"),
    (31091, "targeted_spread"),
    (31471, "targeted_spread"),
    (31806, "targeted_spread"),
    (32146, "targeted_spread"),
    (34081, "targeted_spread"),
    (36108, "targeted_spread"),
    (38043, "targeted_spread"),
    (40093, "targeted_spread"),
    (42093, "targeted_spread"),
    (44008, "targeted_spread"),
    (46008, "targeted_spread"),
    (47943, "targeted_spread"),
    (52083, "targeted_spread"),
    (56003, "targeted_spread"),
    (60073, "targeted_spread"),
    (64148, "targeted_spread"),
    (65248, "targeted_spread"),
    (65608, "targeted_spread"),
    (65923, "targeted_spread"),
    (66263, "targeted_spread"),
    (67273, "targeted_spread"),
    (67613, "targeted_spread"),
    (67973, "targeted_spread"),
    (68288, "targeted_spread"),
    (69323, "targeted_spread"),
    (69658, "targeted_spread"),
    (69998, "targeted_spread"),
    (70313, "targeted_spread"),
    (72223, "targeted_spread"),
    (73283, "targeted_spread"),
    (73618, "targeted_spread"),
    (73958, "targeted_spread"),
    (74298, "targeted_spread"),
    (75283, "targeted_spread"),
    (75668, "targeted_spread"),
    (75983, "targeted_spread"),
    (76323, "targeted_spread"),
    (77333, "targeted_spread"),
    (77693, "targeted_spread"),
    (78058, "targeted_spread"),
    (78393, "targeted_spread"),
    (80373, "targeted_spread"),
    (82418, "targeted_spread"),
    (84513, "targeted_spread"),
    (86468, "targeted_spread"),
    (88543, "targeted_spread"),
    (90613, "targeted_spread"),
    (92613, "targeted_spread"),
    (94618, "targeted_spread"),
    (96688, "targeted_spread"),
    (100713, "targeted_spread"),
    (104858, "targeted_spread"),
    (108973, "targeted_spread"),
    (113048, "targeted_spread"),
    (114148, "targeted_spread"),
    (114508, "targeted_spread"),
    (114823, "targeted_spread"),
    (115188, "targeted_spread"),
    (116178, "targeted_spread"),
    (116558, "targeted_spread"),
    (116873, "targeted_spread"),
    (117188, "targeted_spread"),
    (118223, "targeted_spread"),
    (118583, "targeted_spread"),
    (118943, "targeted_spread"),
    (119258, "targeted_spread")
]

timings_guitar = [
    (700, "sprinkler"),
    (950, "sprinkler"),
    (1105, "sprinkler"),
    (1310, "sprinkler"),
    (1625, "sprinkler"),
    (1940, "sprinkler"),
    (2212, "sprinkler"),
    (2502, "sprinkler"),
    (2842, "sprinkler"),
    (2997, "sprinkler"),
    (3202, "sprinkler"),
    (3537, "sprinkler"),
    (3897, "sprinkler"),
    (4237, "sprinkler"),
    (4552, "sprinkler"),
    (4847, "sprinkler"),
    (5002, "sprinkler"),
    (5202, "sprinkler"),
    (5542, "sprinkler"),
    (5877, "sprinkler"),
    (6172, "sprinkler"),
    (6462, "sprinkler"),
    (6732, "sprinkler"),
    (6892, "sprinkler"),
    (7117, "sprinkler"),
    (7477, "sprinkler"),
    (7747, "sprinkler"),
    (7927, "sprinkler"),
    (8127, "sprinkler"),
    (8467, "sprinkler"),
    (8692, "sprinkler"),
    (8852, "sprinkler"),
    (9077, "sprinkler"),
    (9387, "sprinkler"),
    (9727, "sprinkler"),
    (10042, "sprinkler"),
    (10402, "sprinkler"),
    (10697, "sprinkler"),
    (10877, "sprinkler"),
    (11077, "sprinkler"),
    (11392, "sprinkler"),
    (11732, "sprinkler"),
    (12047, "sprinkler"),
    (12407, "sprinkler"),
    (12677, "sprinkler"),
    (12857, "sprinkler"),
    (13062, "sprinkler"),
    (13372, "sprinkler"),
    (13687, "sprinkler"),
    (14007, "sprinkler"),
    (14367, "sprinkler"),
    (14612, "sprinkler"),
    (14792, "sprinkler"),
    (15017, "sprinkler"),
    (15352, "sprinkler"),
    (15692, "sprinkler"),
    (16027, "sprinkler"),
    (16367, "sprinkler"),
    (16637, "sprinkler"),
    (16817, "sprinkler"),
    (17022, "sprinkler"),
    (17357, "sprinkler"),
    (17652, "sprinkler"),
    (17962, "sprinkler"),
    (18327, "sprinkler"),
    (18552, "sprinkler"),
    (18727, "sprinkler"),
    (18932, "sprinkler"),
    (19272, "sprinkler"),
    (19587, "sprinkler"),
    (19927, "sprinkler"),
    (20262, "sprinkler"),
    (20532, "sprinkler"),
    (20712, "sprinkler"),
    (20912, "sprinkler"),
    (21227, "sprinkler"),
    (21522, "sprinkler"),
    (21837, "sprinkler"),
    (22217, "sprinkler"),
    (22467, "sprinkler"),
    (22647, "sprinkler"),
    (22847, "sprinkler"),
    (23232, "sprinkler"),
    (23477, "sprinkler"),
    (23637, "sprinkler"),
    (23862, "sprinkler"),
    (24197, "sprinkler"),
    (24447, "sprinkler"),
    (24627, "sprinkler"),
    (24827, "sprinkler"),
    (25142, "sprinkler"),
    (25482, "sprinkler"),
    (25797, "sprinkler"),
    (26132, "sprinkler"),
    (26382, "sprinkler"),
    (26537, "sprinkler"),
    (26787, "sprinkler"),
    (27127, "sprinkler"),
    (27462, "sprinkler"),
    (27802, "sprinkler"),
    (28182, "sprinkler"),
    (28407, "sprinkler"),
    (28587, "sprinkler"),
    (28792, "sprinkler"),
    (29172, "sprinkler"),
    (29487, "sprinkler"),
    (29827, "sprinkler"),
    (30162, "sprinkler"),
    (30387, "sprinkler"),
    (30547, "sprinkler"),
    (30792, "sprinkler"),
    (31177, "sprinkler"),
    (31422, "sprinkler"),
    (31582, "sprinkler"),
    (31807, "sprinkler"),
    (32212, "sprinkler"),
    (32412, "sprinkler"),
    (32572, "sprinkler"),
    (32817, "sprinkler"),
    (33182, "sprinkler"),
    (33517, "sprinkler"),
    (33812, "sprinkler"),
    (34212, "sprinkler"),
    (34462, "sprinkler"),
    (34642, "sprinkler"),
    (34867, "sprinkler"),
    (35207, "sprinkler"),
    (35542, "sprinkler"),
    (35857, "sprinkler"),
    (36217, "sprinkler"),
    (36467, "sprinkler"),
    (36692, "sprinkler"),
    (36912, "sprinkler"),
    (37142, "sprinkler"),
    (37477, "sprinkler"),
    (37792, "sprinkler"),
    (38152, "sprinkler"),
    (38422, "sprinkler"),
    (38627, "sprinkler"),
    (38942, "sprinkler"),
    (39164, "sprinkler"),
    (39504, "sprinkler"),
    (39819, "sprinkler"),
    (40179, "sprinkler"),
    (40404, "sprinkler"),
    (40584, "sprinkler"),
    (40829, "sprinkler"),
    (41144, "sprinkler"),
    (41464, "sprinkler"),
    (41824, "sprinkler"),
    (42159, "sprinkler"),
    (42384, "sprinkler"),
    (42564, "sprinkler"),
    (42789, "sprinkler"),
    (43129, "sprinkler"),
    (43441, "sprinkler"),
    (43781, "sprinkler"),
    (44141, "sprinkler"),
    (44366, "sprinkler"),
    (44546, "sprinkler"),
    (44746, "sprinkler"),
    (45086, "sprinkler"),
    (45426, "sprinkler"),
    (45741, "sprinkler"),
    (46101, "sprinkler"),
    (46346, "sprinkler"),
    (46501, "sprinkler"),
    (46706, "sprinkler"),
    (47046, "sprinkler"),
    (47381, "sprinkler"),
    (47696, "sprinkler"),
    (48101, "sprinkler"),
    (48326, "sprinkler"),
    (48506, "sprinkler"),
    (48731, "sprinkler"),
    (49071, "sprinkler"),
    (49406, "sprinkler"),
    (49746, "sprinkler"),
    (50126, "sprinkler"),
    (50376, "sprinkler"),
    (50531, "sprinkler"),
    (50756, "sprinkler"),
    (51121, "sprinkler"),
    (51456, "sprinkler"),
    (51771, "sprinkler"),
    (52106, "sprinkler"),
    (52356, "sprinkler"),
    (52536, "sprinkler"),
    (52781, "sprinkler"),
    (53121, "sprinkler"),
    (53456, "sprinkler"),
    (53771, "sprinkler"),
    (54131, "sprinkler"),
    (54401, "sprinkler"),
    (54561, "sprinkler"),
    (54806, "sprinkler"),
    (55081, "sprinkler"),
    (55461, "sprinkler"),
    (55776, "sprinkler"),
    (56161, "sprinkler"),
    (56381, "sprinkler"),
    (56541, "sprinkler"),
    (56771, "sprinkler"),
    (57126, "sprinkler"),
    (57486, "sprinkler"),
    (57801, "sprinkler"),
    (58141, "sprinkler"),
    (58406, "sprinkler"),
    (58611, "sprinkler"),
    (58816, "sprinkler"),
    (59151, "sprinkler"),
    (59511, "sprinkler"),
    (59851, "sprinkler"),
    (60206, "sprinkler"),
    (60436, "sprinkler"),
    (60616, "sprinkler"),
    (60841, "sprinkler"),
    (61201, "sprinkler"),
    (61516, "sprinkler"),
    (61831, "sprinkler"),
    (62166, "sprinkler"),
    (62436, "sprinkler"),
    (62616, "sprinkler"),
    (62841, "sprinkler"),
    (63181, "sprinkler"),
    (63516, "sprinkler"),
    (63881, "sprinkler"),
    (64306, "sprinkler"),
    (64506, "sprinkler"),
    (64686, "sprinkler"),
    (64911, "sprinkler"),
    (65271, "sprinkler"),
    (65611, "sprinkler"),
    (65926, "sprinkler"),
    (66286, "sprinkler"),
    (66581, "sprinkler"),
    (66761, "sprinkler"),
    (66986, "sprinkler"),
    (67321, "sprinkler"),
    (67636, "sprinkler"),
    (67976, "sprinkler"),
    (68311, "sprinkler"),
    (68581, "sprinkler"),
    (68763, "sprinkler"),
    (68988, "sprinkler"),
    (69303, "sprinkler"),
    (69683, "sprinkler"),
    (69978, "sprinkler"),
    (70313, "sprinkler"),
    (70563, "sprinkler"),
    (70763, "sprinkler"),
    (70968, "sprinkler"),
    (71353, "sprinkler"),
    (71643, "sprinkler"),
    (71983, "sprinkler"),
    (72363, "sprinkler"),
    (72633, "sprinkler"),
    (72793, "sprinkler"),
    (73038, "sprinkler"),
    (73398, "sprinkler"),
    (73688, "sprinkler"),
    (74003, "sprinkler"),
    (74363, "sprinkler"),
    (74613, "sprinkler"),
    (74793, "sprinkler"),
    (75018, "sprinkler"),
    (75378, "sprinkler"),
    (75718, "sprinkler"),
    (76028, "sprinkler"),
    (76413, "sprinkler"),
    (76638, "sprinkler"),
    (76843, "sprinkler"),
    (77023, "sprinkler"),
    (77403, "sprinkler"),
    (77743, "sprinkler"),
    (78058, "sprinkler"),
    (78463, "sprinkler"),
    (78708, "sprinkler"),
    (78868, "sprinkler"),
    (79093, "sprinkler"),
    (79498, "sprinkler"),
    (79743, "sprinkler"),
    (79923, "sprinkler"),
    (80128, "sprinkler"),
    (80533, "sprinkler"),
    (80758, "sprinkler"),
    (80918, "sprinkler"),
    (81143, "sprinkler"),
    (81528, "sprinkler"),
    (81863, "sprinkler"),
    (82198, "sprinkler"),
    (82580, "sprinkler"),
    (82810, "sprinkler"),
    (82985, "sprinkler"),
    (83210, "sprinkler"),
    (83595, "sprinkler"),
    (83935, "sprinkler"),
    (84225, "sprinkler"),
    (84650, "sprinkler"),
    (84875, "sprinkler"),
    (85035, "sprinkler"),
    (85285, "sprinkler"),
    (85620, "sprinkler"),
    (85845, "sprinkler"),
    (86025, "sprinkler"),
    (86270, "sprinkler"),
    (86675, "sprinkler"),
    (86880, "sprinkler"),
    (87040, "sprinkler"),
    (87285, "sprinkler"),
    (87625, "sprinkler"),
    (87940, "sprinkler"),
    (88320, "sprinkler"),
    (88705, "sprinkler"),
    (88930, "sprinkler"),
    (89110, "sprinkler"),
    (89335, "sprinkler"),
    (89715, "sprinkler"),
    (90010, "sprinkler"),
    (90325, "sprinkler"),
    (90730, "sprinkler"),
    (90930, "sprinkler"),
    (91110, "sprinkler"),
    (91315, "sprinkler"),
    (91740, "sprinkler"),
    (92010, "sprinkler"),
    (92350, "sprinkler"),
    (92755, "sprinkler"),
    (92935, "sprinkler"),
    (93140, "sprinkler"),
    (93385, "sprinkler"),
    (93745, "sprinkler"),
    (93970, "sprinkler"),
    (94150, "sprinkler"),
    (94355, "sprinkler"),
    (94755, "sprinkler"),
    (95005, "sprinkler"),
    (95205, "sprinkler"),
    (95410, "sprinkler"),
    (95770, "sprinkler"),
    (96110, "sprinkler"),
    (96445, "sprinkler"),
    (96850, "sprinkler"),
    (97075, "sprinkler"),
    (97255, "sprinkler"),
    (97480, "sprinkler"),
    (97795, "sprinkler"),
    (98160, "sprinkler"),
    (98475, "sprinkler"),
    (98810, "sprinkler"),
    (99080, "sprinkler"),
    (99260, "sprinkler"),
    (99465, "sprinkler"),
    (99845, "sprinkler"),
    (100160, "sprinkler"),
    (100495, "sprinkler"),
    (100860, "sprinkler"),
    (101125, "sprinkler"),
    (101285, "sprinkler"),
    (101510, "sprinkler"),
    (101890, "sprinkler"),
    (102210, "sprinkler"),
    (102525, "sprinkler"),
    (102925, "sprinkler"),
    (103175, "sprinkler"),
    (103355, "sprinkler"),
    (103600, "sprinkler"),
    (103915, "sprinkler"),
    (104215, "sprinkler"),
    (104390, "sprinkler"),
    (104590, "sprinkler"),
    (105000, "sprinkler"),
    (105245, "sprinkler"),
    (105425, "sprinkler"),
    (105630, "sprinkler"),
    (105965, "sprinkler"),
    (106325, "sprinkler"),
    (106685, "sprinkler"),
    (107025, "sprinkler"),
    (107295, "sprinkler"),
    (107450, "sprinkler"),
    (107675, "sprinkler"),
    (108035, "sprinkler"),
    (108350, "sprinkler"),
    (108690, "sprinkler"),
    (109050, "sprinkler"),
    (109320, "sprinkler"),
    (109500, "sprinkler"),
    (109705, "sprinkler"),
    (110040, "sprinkler"),
    (110420, "sprinkler"),
    (110760, "sprinkler"),
    (111120, "sprinkler"),
    (111365, "sprinkler"),
    (111570, "sprinkler"),
    (111820, "sprinkler"),
    (112180, "sprinkler"),
    (112425, "sprinkler"),
    (112605, "sprinkler"),
    (112805, "sprinkler"),
    (113190, "sprinkler"),
    (113415, "sprinkler"),
    (113615, "sprinkler"),
    (113800, "sprinkler"),
    (114180, "sprinkler"),
    (114520, "sprinkler"),
    (114855, "sprinkler"),
    (115235, "sprinkler"),
    (115485, "sprinkler"),
    (115665, "sprinkler"),
    (115870, "sprinkler"),
    (116230, "sprinkler"),
    (116545, "sprinkler"),
    (116860, "sprinkler"),
    (117240, "sprinkler"),
    (117510, "sprinkler"),
    (117670, "sprinkler"),
    (117875, "sprinkler"),
    (118255, "sprinkler"),
    (118615, "sprinkler"),
    (118930, "sprinkler"),
    (119490, "sprinkler")
]

timings_bass = [
    (8322, "homing"),
    (9336, "homing"),
    (10209, "homing"),
    (10303, "homing"),
    (11255, "homing"),
    (12081, "homing"),
    (12237, "homing"),
    (13189, "homing"),
    (13969, "homing"),
    (14125, "homing"),
    (15186, "homing"),
    (16215, "homing"),
    (17214, "homing"),
    (18197, "homing"),
    (19195, "homing"),
    (20053, "homing"),
    (20147, "homing"),
    (21129, "homing"),
    (21473, "homing"),
    (21785, "homing"),
    (22050, "homing"),
    (22908, "homing"),
    (23064, "homing"),
    (23407, "homing"),
    (23766, "homing"),
    (24062, "homing"),
    (25061, "homing"),
    (25872, "homing"),
    (26028, "homing"),
    (27026, "homing"),
    (28009, "homing"),
    (29039, "homing"),
    (30022, "homing"),
    (31020, "homing"),
    (32018, "homing"),
    (32908, "homing"),
    (33282, "homing"),
    (33984, "homing"),
    (35092, "homing"),
    (36059, "homing"),
    (37026, "homing"),
    (37822, "homing"),
    (38960, "homing"),
    (39974, "homing"),
    (41004, "homing"),
    (42034, "homing"),
    (43079, "homing"),
    (43999, "homing"),
    (44935, "homing"),
    (45934, "homing"),
    (46916, "homing"),
    (47634, "homing"),
    (47962, "homing"),
    (48898, "homing"),
    (49958, "homing"),
    (50068, "homing"),
    (50926, "homing"),
    (51955, "homing"),
    (52049, "homing"),
    (52969, "homing"),
    (53328, "homing"),
    (53702, "homing"),
    (54014, "homing"),
    (54997, "homing"),
    (55340, "homing"),
    (55652, "homing"),
    (56011, "homing"),
    (57025, "homing"),
    (57883, "homing"),
    (58086, "homing"),
    (58850, "homing"),
    (59053, "homing"),
    (60020, "homing"),
    (60941, "homing"),
    (61128, "homing"),
    (61736, "homing"),
    (62095, "homing"),
    (62922, "homing"),
    (63109, "homing"),
    (63468, "homing"),
    (63780, "homing"),
    (64123, "homing"),
    (65168, "homing"),
    (66026, "homing"),
    (66167, "homing"),
    (67150, "homing"),
    (67992, "homing"),
    (68117, "homing"),
    (69224, "homing"),
    (69536, "homing"),
    (69864, "homing"),
    (70176, "homing"),
    (71268, "homing"),
    (71596, "homing"),
    (71876, "homing"),
    (72173, "homing"),
    (73202, "homing"),
    (73904, "homing"),
    (74216, "homing"),
    (75074, "homing"),
    (75230, "homing"),
    (75605, "homing"),
    (75932, "homing"),
    (76276, "homing"),
    (77321, "homing"),
    (77945, "homing"),
    (78319, "homing"),
    (79286, "homing"),
    (79661, "homing"),
    (80004, "homing"),
    (80316, "homing"),
    (81112, "homing"),
    (81424, "homing"),
    (81736, "homing"),
    (82032, "homing"),
    (82328, "homing"),
    (83358, "homing"),
    (84029, "homing"),
    (84388, "homing"),
    (85402, "homing"),
    (86244, "homing"),
    (86603, "homing"),
    (87523, "homing"),
    (88475, "homing"),
    (89520, "homing"),
    (90534, "homing"),
    (91548, "homing"),
    (92515, "homing"),
    (93576, "homing"),
    (94216, "homing"),
    (94574, "homing"),
    (95573, "homing"),
    (96275, "homing"),
    (96618, "homing"),
    (97538, "homing"),
    (98693, "homing"),
    (99582, "homing"),
    (100549, "homing"),
    (100705, "homing"),
    (101657, "homing"),
    (102000, "homing"),
    (102359, "homing"),
    (102718, "homing"),
    (103778, "homing"),
    (104059, "homing"),
    (104434, "homing"),
    (104746, "homing"),
    (105728, "homing"),
    (106118, "homing"),
    (106493, "homing"),
    (106820, "homing"),
    (107850, "homing"),
    (108770, "homing"),
    (109894, "homing"),
    (110564, "homing"),
    (110923, "homing"),
    (111953, "homing"),
    (112265, "homing"),
    (112624, "homing"),
    (112982, "homing"),
    (114012, "homing"),
    (115042, "homing"),
    (115946, "homing"),
    (116102, "homing"),
    (117023, "homing"),
    (117974, "homing"),
    (118427, "homing"),
    (118801, "homing"),
    (119160, "homing"),
]
