"""Implements a simple two-dimensional vector class."""

import math

class Vector2d(object):
    """Two dimensional vector, with overloaded operators.
    
    Since we want to be able to use instances of this class as dictionary keys,
    the fields of this class (x and y) should be treated as immutable. Don't
    modify them!
    """
    def __init__(self, x, y):
        self.x = x
        self.y = y
    def __repr__(self):
        return "Vector2d(%r, %r)" % (self.x, self.y)
    def __eq__(self, other):
        return self.x == other.x and self.y == other.y
    def __add__(self, other):
        return Vector2d(self.x + other.x, self.y + other.y)
    def __sub__(self, other):
        return Vector2d(self.x - other.x, self.y - other.y)
    def __mul__(self, other):
        return Vector2d(self.x * other, self.y * other)
    def __rmul__(self, other):
        return Vector2d(self.x * other, self.y * other)
    def __div__(self, other):
        other = float(other)
        return Vector2d(self.x / other, self.y / other)
    def __hash__(self):
        """Calculates a hash code for this vector using a tuple."""
        return (self.x, self.y).__hash__()

    def coords(self):
        return (self.x, self.y)

    def magnitude_squared(self):
        """Returns the square of the vector's magnitude.
        
        This is much faster than magnitude(), and can be used to optimize a
        condition like (v.magnitude() < some_number) to 
        (v.magnitude_squared() < some_number**2)
        """
        return self.dot(self)

    def magnitude(self):
        """Returns the magnitude of the vector.
        
        This is very slow since it does a square root. Try to use
        magnitude_squared() instead where possible.
        """
        return math.sqrt(self.magnitude_squared())

    def dot(self, other):
        """Returns the dot product of this vector and another one."""
        return self.x * other.x + self.y * other.y

    def cross(self, other):
        """Returns the z-component of the cross product of the two vectors"""
        return self.x * other.y - other.x * self.y

    def project(self, other):
        """Returns self projected onto other."""
        return float(self.dot(other))/other.dot(other)*other

    def is_zero(self):
        """Returns true if you cannot normalize this vector."""
        return self.x**2 + self.y**2 == 0

    def get_normalized(self):
        """Returns a normalized version of the current vector."""
        length = self.magnitude()
        assert(length != 0)
        return Vector2d(self.x / length, self.y / length)
        
    def rotate(self, angle):
        """Returns a rotate version of the current vector, may return very small values"""
        sin_a = math.sin(angle)
        cos_a = math.cos(angle)
        return Vector2d(self.x * cos_a - self.y * sin_a, self.y * cos_a + self.x * sin_a)

    def rotate90(self):
        return Vector2d(-self.y, self.x)
        
    def approx(self, other):
        return (self - other).magnitude_squared() < 0.00001
        
    # Implement more operators when you need them.
    # Make sure to add tests though =p. Nothing hurts more than debugging for
    # hours before finding out there's a typo in one of these functions >_>
        
# Constants. (should not be modified... is this pythonesque?)

Vector2d.NORTH = Vector2d.UP    = Vector2d(0, -1)
Vector2d.EAST  = Vector2d.RIGHT = Vector2d(1, 0)
Vector2d.SOUTH = Vector2d.DOWN  = Vector2d(0, 1)
Vector2d.WEST  = Vector2d.LEFT  = Vector2d(-1, 0)

Vector2d.NORTH_EAST = Vector2d.NORTH + Vector2d.EAST
Vector2d.NORTH_WEST = Vector2d.NORTH + Vector2d.WEST
Vector2d.SOUTH_EAST = Vector2d.SOUTH + Vector2d.EAST
Vector2d.SOUTH_WEST = Vector2d.SOUTH + Vector2d.WEST

Vector2d.ZERO = Vector2d(0, 0)

# Tests
if __name__ == "__main__":
    assert(Vector2d(1, 2) + Vector2d(3, 4) == Vector2d(4, 6))
    assert(Vector2d(1, 2) - Vector2d(3, 4) == Vector2d(-2, -2))
    assert(Vector2d(1, 2) * 10 == Vector2d(10, 20))
    assert(Vector2d(1, 2) / 2 == Vector2d(0.5, 1))
    assert(10 * Vector2d(1, 2) == Vector2d(10, 20))
    assert(Vector2d(3, 4).get_normalized() == Vector2d(3/5.0, 4/5.0))
    assert(Vector2d(3, 4).magnitude_squared() == 25)
    assert(Vector2d(6, 8).magnitude() == 10)
    assert(Vector2d(1,0).rotate(math.pi).approx(Vector2d(-1,0)))
    assert(Vector2d(1,1).rotate90() == Vector2d(-1,1))
    assert(Vector2d.NORTH.dot(Vector2d.NORTH) == 1)
    assert(Vector2d.NORTH.dot(Vector2d.WEST) == 0)
    assert(Vector2d(1, 0).dot(Vector2d(2, 2)) == 2)
    assert(Vector2d(1, 0).cross(Vector2d(0, 1)) == 1)
    assert(Vector2d(1, 1).cross(Vector2d(1, -1)) < 0)
    assert(Vector2d(1, 1).project(Vector2d(10, 0)) == Vector2d(1, 0))
    assert(Vector2d(4, 6).project(Vector2d(5, 1)) == Vector2d(5, 1))
    assert(Vector2d.ZERO.is_zero())
    assert(not Vector2d.SOUTH.is_zero())
